/*
 * DMA support for the JZ4730.
 *
 * Copyright (C) 2021, 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/sys/types.h>
#include <stdint.h>



/* Enumerated types for various transfer parameters. */

enum Dma_jz4730_request_type
{
  Dma_request_external        = 0,
  Dma_request_pcmcia_out      = 4,
  Dma_request_pcmcia_in       = 5,
  Dma_request_auto            = 8,
  Dma_request_des_out         = 10,
  Dma_request_des_in          = 11,
  Dma_request_uart3_out       = 14,
  Dma_request_uart3_in        = 15,
  Dma_request_uart2_out       = 16,
  Dma_request_uart2_in        = 17,
  Dma_request_uart1_out       = 18,
  Dma_request_uart1_in        = 19,
  Dma_request_uart0_out       = 20,
  Dma_request_uart0_in        = 21,
  Dma_request_ssi_out         = 22,
  Dma_request_ssi_in          = 23,
  Dma_request_aic_out         = 24,
  Dma_request_aic_in          = 25,
  Dma_request_msc_out         = 26,
  Dma_request_msc_in          = 27,
  Dma_request_ost2_underflow  = 28,
};

enum Dma_jz4730_ext_level
{
  Dma_ext_active_high                   = 0,
  Dma_ext_active_low                    = 1,
};

enum Dma_jz4730_ext_output_mode_cycle
{
  Dma_ext_output_mode_read_cycle        = 0,
  Dma_ext_output_mode_write_cycle       = 1,
};

enum Dma_jz4730_ext_req_detect_mode
{
  Dma_ext_req_detect_mode_low_level     = 0,
  Dma_ext_req_detect_mode_falling_edge  = 1,
  Dma_ext_req_detect_mode_high_level    = 2,
  Dma_ext_req_detect_mode_rising_edge   = 3,
};



#ifdef __cplusplus

#include <l4/devices/dma-generic.h>
#include <l4/devices/hw_mmio_register_block.h>

// Forward declaration.

class Dma_jz4730_chip;



// DMA channel.

class Dma_jz4730_channel : public Dma_channel
{
private:
  Hw::Register_block<32> _regs;
  Dma_chip *_chip;
  uint8_t _channel;
  l4_cap_idx_t _irq = L4_INVALID_CAP;

  // External transfer properties with defaults.

  enum Dma_jz4730_ext_level _ext_output_polarity = Dma_ext_active_high;
  enum Dma_jz4730_ext_level _ext_end_of_process_mode = Dma_ext_active_high;
  enum Dma_jz4730_ext_output_mode_cycle _ext_output_mode_cycle = Dma_ext_output_mode_read_cycle;
  enum Dma_jz4730_ext_req_detect_mode _ext_req_detect_mode = Dma_ext_req_detect_mode_high_level;

public:
  Dma_jz4730_channel(Dma_chip *chip, uint8_t channel, l4_addr_t start, l4_cap_idx_t irq);

  unsigned int transfer(uint32_t source, uint32_t destination,
                        unsigned int count,
                        bool source_increment, bool destination_increment,
                        uint8_t source_width, uint8_t destination_width,
                        uint8_t transfer_unit_size,
                        int type=Dma_request_auto,
                        l4_addr_t desc_vaddr = 0,
                        l4re_dma_space_dma_addr_t desc_paddr = 0);

  unsigned int wait();

  // External transfer property configuration.

  void set_output_polarity(enum Dma_jz4730_ext_level polarity)
  { _ext_output_polarity = polarity; }

  void set_end_of_process_mode(enum Dma_jz4730_ext_level mode)
  { _ext_end_of_process_mode = mode; }

  void set_output_mode_cycle(enum Dma_jz4730_ext_output_mode_cycle cycle)
  { _ext_output_mode_cycle = cycle; }

  void set_req_detect_mode(enum Dma_jz4730_ext_req_detect_mode mode)
  { _ext_req_detect_mode = mode; }

protected:
  // Transfer property configuration.

  uint32_t encode_external_transfer(enum Dma_jz4730_request_type type);

  uint32_t encode_req_detect_int_length(uint8_t units);

  uint32_t encode_source_port_width(uint8_t width);

  uint32_t encode_destination_port_width(uint8_t width);

  uint32_t encode_transfer_unit_size(uint8_t size);

  // Transaction control.

  void ack_irq();

  void clear_errors();

  bool completed();

  bool error();

  bool halted();

  bool wait_for_irq();

  bool wait_for_irq(unsigned int timeout);
};

// DMA device control.

class Dma_jz4730_chip : public Dma_chip
{
private:
  Hw::Register_block<32> _regs;
  l4_addr_t _start, _end;
  Cpm_chip *_cpm;

public:
  Dma_jz4730_chip(l4_addr_t start, l4_addr_t end, Cpm_chip *cpm);

  void disable();

  void enable();

  Dma_channel *get_channel(uint8_t channel, l4_cap_idx_t irq = L4_INVALID_CAP);

  bool have_interrupt(uint8_t channel);
};

Dma_chip *jz4730_dma_chip(l4_addr_t start, l4_addr_t end, Cpm_chip *cpm);

#endif /* __cplusplus */



/* C language interface. */

EXTERN_C_BEGIN

void *jz4730_dma_init(l4_addr_t start, l4_addr_t end, void *cpm);

void jz4730_dma_disable(void *dma_chip);

void jz4730_dma_enable(void *dma_chip);

void *jz4730_dma_get_channel(void *dma, uint8_t channel, l4_cap_idx_t irq);

void jz4730_dma_set_output_polarity(void *dma_channel, enum Dma_jz4730_ext_level polarity);

void jz4730_dma_set_end_of_process_mode(void *dma_channel, enum Dma_jz4730_ext_level mode);

void jz4730_dma_set_output_mode_cycle(void *dma_channel, enum Dma_jz4730_ext_output_mode_cycle cycle);

void jz4730_dma_set_req_detect_mode(void *dma_channel, enum Dma_jz4730_ext_req_detect_mode mode);

unsigned int jz4730_dma_transfer(void *dma_channel,
                                 uint32_t source, uint32_t destination,
                                 unsigned int count,
                                 int source_increment, int destination_increment,
                                 uint8_t source_width, uint8_t destination_width,
                                 uint8_t transfer_unit_size,
                                 enum Dma_jz4730_request_type type);

unsigned int jz4730_dma_wait(void *dma_channel);

EXTERN_C_END
