/*
 * MSC (MMC/SD controller) peripheral support.
 *
 * Copyright (C) 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/err.h>

#include "msc-defs.h"
#include "msc-x1600.h"



// Channel abstraction.

Msc_x1600_channel::Msc_x1600_channel(l4_addr_t msc_start, l4_addr_t addr,
                                     l4_cap_idx_t irq,
                                     Cpm_chip *cpm, enum Clock_identifiers clock,
                                     Dma_channel *dma,
                                     enum Dma_x1600_request_type request_type_in,
                                     enum Dma_x1600_request_type request_type_out)
: Msc_channel(msc_start, addr, irq, cpm, clock),
  _dma(dma),
  _request_type_in(request_type_in),
  _request_type_out(request_type_out)
{
}

// Reset for X1600 and other recent SoCs only.

void
Msc_x1600_channel::reset()
{
  _regs[Msc_control] = _regs[Msc_control] | Control_reset;

  // NOTE: On the X1600, Msc_control tends to read as zero, anyway, making this
  // NOTE: equivalent to clearing the register.

  _regs[Msc_control] = _regs[Msc_control] & ~Control_reset;

  // Sufficient for other SoCs...

  while (_regs[Msc_status] & Status_resetting);
}

// Request the transfer of the indicated number of bytes between two physical
// addresses in the indicated direction, returning the number of bytes to be
// transferred.
// NOTE: To be consolidated into a generic method that uses generic request types.

uint32_t
Msc_x1600_channel::transfer(l4re_dma_space_dma_addr_t from_paddr,
                            l4re_dma_space_dma_addr_t to_paddr,
                            bool recv, uint32_t count)
{
  uint32_t unit_size = 32;
  uint32_t unit_count = count / unit_size;
  uint32_t to_transfer;

  to_transfer = _dma->transfer(from_paddr, to_paddr,
                               unit_count,
                               recv ? false : true, // increment source if sending
                               recv ? true : false, // increment destination if receiving
                               4, 4, unit_size,
                               (int) (recv ? _request_type_in : _request_type_out));

  return to_transfer * unit_size;
}



// Peripheral abstraction.

Msc_x1600_chip::Msc_x1600_chip(l4_addr_t msc_start, l4_addr_t start,
                               l4_addr_t end, Cpm_chip *cpm)
: Msc_chip(msc_start, start, end, cpm)
{
}

Msc_channel *Msc_x1600_chip::get_channel(uint8_t channel, l4_cap_idx_t irq,
                                         Dma_channel *dma)
{
  if (channel < num_channels())
    return new Msc_x1600_channel(_msc_start + channel * Msc_channel_offset,
                                 _start + channel * Msc_channel_offset,
                                 irq, _cpm, _clocks[channel], dma,
                                 _in_types[channel], _out_types[channel]);
  else
    throw -L4_EINVAL;
}

Msc_chip *x1600_msc_chip(l4_addr_t msc_start, l4_addr_t start, l4_addr_t end,
                         Cpm_chip *cpm)
{
  return new Msc_x1600_chip(msc_start, start, end, cpm);
}



// C language interface functions.

void *x1600_msc_init(l4_addr_t msc_start, l4_addr_t start, l4_addr_t end,
                     void *cpm)
{
  return (void *) x1600_msc_chip(msc_start, start, end,
    static_cast<Cpm_chip *>(cpm));
}

void *x1600_msc_get_channel(void *msc, uint8_t channel, l4_cap_idx_t irq,
                            void *dma)
{
  return static_cast<Msc_chip *>(msc)->get_channel(channel, irq,
    static_cast<Dma_channel *>(dma));
}

struct msc_card *x1600_msc_get_cards(void *msc_channel)
{
  return static_cast<Msc_channel *>(msc_channel)->get_cards();
}

uint8_t x1600_msc_num_cards(void *msc_channel)
{
  return static_cast<Msc_channel *>(msc_channel)->num_cards();
}

void x1600_msc_enable(void *msc_channel)
{
  static_cast<Msc_channel *>(msc_channel)->enable();
}

uint32_t x1600_msc_read(void *msc_channel, uint8_t card,
                        struct dma_region *region,
                        uint64_t address, uint64_t count)
{
  return static_cast<Msc_channel *>(msc_channel)->read(card, region,
    address, count);
}

uint32_t x1600_msc_read_blocks(void *msc_channel, uint8_t card,
                               struct dma_region *region,
                               uint32_t block_address, uint32_t block_count)
{
  return static_cast<Msc_channel *>(msc_channel)->read_blocks(card, region,
    block_address, block_count);
}
